﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace FFT_Template
{
    public partial class FourierDemo : Form
    {
        Bitmap image, processed, display;
        string image_path;
        string filter = "Image | *.png; *.jpg; *.gif";
        OpenFileDialog ofd;
        SaveFileDialog sfd;
        public static int Size;
        public static int og_width, og_height;

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                image_path = ofd.FileName;
                image = (Bitmap)Bitmap.FromFile(image_path);
                display = image.ScaleImage(image_frame.Width, image_frame.Height);
                processed = image.Add_Padding();
                image_frame.Image = display;
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sfd.FileName = image_path;
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                processed.Save(sfd.FileName);
            }
        }


        private void idealBandrejectFilterToolStripMenuItem_Click(object sender, EventArgs e)
        {
            float[][][] normalized = processed.Normalize_Image();
            converted = new Complex[3][][];

            for (int i = 0; i < 3; i++)
            {
                converted[i] = FFT.Forward(normalized, i);
            }

            Complex[][][] filtered = Filters.IdealBandrejectFilter(converted, 30, 8);
            var tmp = FFT.InverseToFloat(filtered);
            processed = tmp.ToImage();
            processed = processed.RemovePadding();
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void butterworthBandrejectFilterToolStripMenuItem_Click(object sender, EventArgs e)
        {
            float[][][] normalized = processed.Normalize_Image();
            converted = new Complex[3][][];

            for (int i = 0; i < 3; i++)
            {
                converted[i] = FFT.Forward(normalized, i);
            }

            Complex[][][] filtered = Filters.ButterworthBandrejectFilter(converted, 60, 2, 16);
            var tmp = FFT.InverseToFloat(filtered);
            processed = tmp.ToImage();
            processed = processed.RemovePadding();
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void gaussianBandrejectFilterToolStripMenuItem_Click(object sender, EventArgs e)
        {
            float[][][] normalized = processed.Normalize_Image();
            converted = new Complex[3][][];

            for (int i = 0; i < 3; i++)
            {
                converted[i] = FFT.Forward(normalized, i);
            }

            Complex[][][] filtered = Filters.GaussianBandrejectFilter(converted, 60, 16);
            var tmp = FFT.InverseToFloat(filtered);
            processed = tmp.ToImage();
            processed = processed.RemovePadding();
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void visualizeFourierToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Bitmap fourier = converted[1].VisualizeFourier();
            image_frame.Image = fourier.ScaleImage(image_frame.Width, image_frame.Height);
            processed = fourier;
        }

        public static int padding_top, padding_left;
        public static Complex[][][] converted;

        public FourierDemo()
        {
            InitializeComponent();
            ofd = new OpenFileDialog();
            sfd = new SaveFileDialog();
            ofd.Filter = filter;
            sfd.Filter = filter;
        }
    }

    static class Ext
    {
        public static Bitmap VisualizeFourier(this Complex[][] f)
        {
            float[] frequencies = new float[FFT.Size * FFT.Size];
            for (int i = 0; i < f.Length; i++)
            {
                for (int j = 0; j < f[0].Length; j++)
                {
                    frequencies[j * f.Length + i] = Complex.Modulus(f[j][i]);
                }
            }

            float min = frequencies.Min();
            float max = frequencies.Max();
            double log_constant = 255 / Math.Log10(1 + 255);
            for (int i = 0; i < frequencies.Length; i++)
            {
                frequencies[i] = 255 * (frequencies[i] - min) / (max - min);
                frequencies[i] = (int)(Math.Log10(1 + frequencies[i]) * log_constant);
            }

            Bitmap image = new Bitmap(FFT.Size, FFT.Size);
            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, FFT.Size, FFT.Size),
                ImageLockMode.WriteOnly,
                PixelFormat.Format32bppArgb);
            int bytes = image_data.Stride * image_data.Height;
            byte[] result = new byte[bytes];
            for (int y = 0; y < FFT.Size; y++)
            {
                for (int x = 0; x < FFT.Size; x++)
                {
                    int pixel_position = y * image_data.Stride + x * 4;
                    for (int i = 0; i < 3; i++)
                    {
                        result[pixel_position + i] = (byte)frequencies[y * FFT.Size + x];
                    }
                    result[pixel_position + 3] = 255;
                }
            }
            Marshal.Copy(result, 0, image_data.Scan0, bytes);
            image.UnlockBits(image_data);
            return image;
        }

        public static Bitmap ScaleImage(this Bitmap image, int width, int height)
        {
            float xratio = (float)width / image.Width;
            float yratio = (float)height / image.Height;
            float ratio = Math.Min(xratio, yratio);

            int w = (int)(image.Width * ratio);
            int h = (int)(image.Height * ratio);

            Bitmap result = new Bitmap(image, new Size(w, h));
            return result;
        }

        public static Bitmap Add_Padding(this Bitmap image)
        {
            int w = image.Width;
            int h = image.Height;
            FourierDemo.og_width = w;
            FourierDemo.og_height = h;
            int n = 0;
            while (FourierDemo.Size <= Math.Max(w, h))
            {
                FourierDemo.Size = (int)Math.Pow(2, n);
                if (FourierDemo.Size == Math.Max(w, h))
                {
                    break;
                }
                n++;
            }
            double horizontal_padding = FourierDemo.Size - w;
            double vertical_padding = FourierDemo.Size - h;
            int left_padding = (int)Math.Floor(horizontal_padding / 2);
            int top_padding = (int)Math.Floor(vertical_padding / 2);
            FourierDemo.padding_left = left_padding;
            FourierDemo.padding_top = top_padding;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format32bppArgb);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];
            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            Bitmap padded_image = new Bitmap(FourierDemo.Size, FourierDemo.Size);

            BitmapData padded_data = padded_image.LockBits(
                new Rectangle(0, 0, FourierDemo.Size, FourierDemo.Size),
                ImageLockMode.WriteOnly,
                PixelFormat.Format32bppArgb);

            int padded_bytes = padded_data.Stride * padded_data.Height;
            byte[] result = new byte[padded_bytes];

            for (int i = 3; i < padded_bytes; i += 4)
            {
                result[i] = 255;
            }

            for (int y = 0; y < h; y++)
            {
                for (int x = 0; x < w; x++)
                {
                    int image_position = y * image_data.Stride + x * 4;
                    int padding_position = y * padded_data.Stride + x * 4;
                    for (int i = 0; i < 3; i++)
                    {
                        result[padded_data.Stride * top_padding + 4 * left_padding + padding_position + i] = buffer[image_position + i];
                    }
                }
            }

            Marshal.Copy(result, 0, padded_data.Scan0, padded_bytes);
            padded_image.UnlockBits(padded_data);

            return padded_image;
        }

        public static Bitmap RemovePadding(this Bitmap image)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format32bppArgb);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];

            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            Bitmap unpadded_image = new Bitmap(FourierDemo.og_width, FourierDemo.og_height);

            BitmapData unpadded_data = unpadded_image.LockBits(
                new Rectangle(0, 0, FourierDemo.og_width, FourierDemo.og_height),
                ImageLockMode.WriteOnly,
                PixelFormat.Format32bppArgb);

            int unpadded_bytes = unpadded_data.Stride * unpadded_data.Height;
            byte[] result = new byte[unpadded_bytes];

            for (int y = 0; y < FourierDemo.og_height; y++)
            {
                for (int x = 0; x < FourierDemo.og_width; x++)
                {
                    int padded_position = y * image_data.Stride + x * 4;
                    int unpadded_position = y * unpadded_data.Stride + x * 4;
                    for (int i = 0; i < 4; i++)
                    {
                        result[unpadded_position + i] = buffer[image_data.Stride * FourierDemo.padding_top + 4 * FourierDemo.padding_left + padded_position + i];
                    }
                }
            }

            Marshal.Copy(result, 0, unpadded_data.Scan0, unpadded_bytes);
            unpadded_image.UnlockBits(unpadded_data);

            return unpadded_image;
        }

        public static float[][][] Normalize_Image(this Bitmap image)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb);
            int bytes = image_data.Stride * image_data.Height;
            float[][][] result = new float[3][][];
            byte[] buffer = new byte[bytes];
            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            int max = buffer.Max();
            int min = buffer.Min();

            int pixel_position;

            for (int c = 0; c < 3; c++)
            {
                result[c] = new float[w][];
                for (int x = 0; x < w; x++)
                {
                    result[c][x] = new float[h];
                    for (int y = 0; y < h; y++)
                    {
                        pixel_position = x * 3 + y * image_data.Stride;
                        result[c][x][y] = ((float)buffer[pixel_position + c] - min) / (max - min);
                    }
                }
            }
            return result;
        }

        public static float[][][] Scale_Data(this float[][][] data)
        {
            float max = 0f;
            float[][][] result = new float[3][][];

            for (int i = 0; i < 3; i++)
            {
                for (int j = 0; j < FFT.Size; j++)
                {
                    for (int k = 0; k < FFT.Size; k++)
                    {
                        max = Math.Max(data[i][j][k], max);
                    }
                }
            }

            for (int i = 0; i < 3; i++)
            {
                result[i] = new float[FFT.Size][];
                for (int j = 0; j < FFT.Size; j++)
                {
                    result[i][j] = new float[FFT.Size];
                    for (int k = 0; k < FFT.Size; k++)
                    {
                        result[i][j][k] = data[i][j][k] / max;
                    }
                }
            }

            return result;
        }

        public static float[][][] Add(this float[][][] a, float[][][] b)
        {
            int c = -1;
            float[][][] result = new float[3][][];
            for (int i = 0; i < 3; i++)
            {
                result[i] = new float[FFT.Size][];
                for (int j = 0; j < FFT.Size; j++)
                {
                    result[i][j] = new float[FFT.Size];
                    for (int k = 0; k < FFT.Size; k++)
                    {
                        result[i][j][k] = a[i][j][k] + c * b[i][j][k];
                    }
                }
            }

            return result;
        }

        public static Bitmap ToImage(this float[][][] data)
        {
            int s = FFT.Size;
            Bitmap result = new Bitmap(s, s);
            BitmapData img_data = result.LockBits(
                new Rectangle(0, 0, s, s),
                ImageLockMode.WriteOnly,
                PixelFormat.Format24bppRgb);
            int bytes = img_data.Stride * img_data.Height;
            byte[] buffer = new byte[bytes];
            float min = 255f;
            float max = 0f;

            for (int i = 0; i < 3; i++)
            {
                for (int j = 0; j < FFT.Size; j++)
                {
                    for (int k = 0; k < FFT.Size; k++)
                    {
                        min = Math.Min(min, data[i][j][k]);
                        max = Math.Max(max, data[i][j][k]);
                    }
                }
            }

            for (int c = 0; c < 3; c++)
            {
                for (int x = 0; x < FFT.Size; x++)
                {
                    for (int y = 0; y < FFT.Size; y++)
                    {
                        int pixel_position = x * 3 + y * img_data.Stride;
                        buffer[pixel_position + c] = (byte)(255 * (data[c][x][y] - min) / (max - min));
                    }
                }
            }

            Marshal.Copy(buffer, 0, img_data.Scan0, bytes);
            result.UnlockBits(img_data);
            return result;
        }
    }
}
